//DIGIKSM.C by Ken Silverman (http://advsys.net/ken)
#include <string.h>
#include <conio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <stdio.h>
#include <math.h>

#define SBBUFSIZ 16384
extern char sbbuf[SBBUFSIZ];
extern volatile long sbplc, sbend;
extern long samplerate, numspeakers, bytespersample, intspersec, halfbufbytes;
extern void initsb (char, long, char, char, char);
extern void changesamplerate (long);
extern void uninitsb ();

extern float lvol[9], rvol[9];
extern long lplc[9], rplc[9];
void adlibinit(long samplerate, long numspeakers, long bytespersample);
void adlib0(long index, long value);
void adlibgetsample(void *sndptr, long numsamples);

static unsigned long calccycles = 1, freecycles = 1;
static long t0, t1;

long gettime586 ();
#pragma aux gettime586 =\
	"dw 0x310f"\
	modify exact [eax edx]

void readmousexy (short *, short *);
#pragma aux readmousexy =\
	"mov ax, 11d"\
	"int 33h"\
	"mov [esi], cx"\
	"mov [edi], dx"\
	parm [esi][edi] modify exact [eax ebx ecx edx]

void readmousebstatus (short *);
#pragma aux readmousebstatus =\
	"mov ax, 5d"\
	"int 33h"\
	"mov [esi], ax"\
	parm [esi] modify exact [eax ebx ecx edx]

unsigned long note[8192], musicstatus, count, countstop;
unsigned int nownote, numnotes, speed, numchans, firstime = 1;
static unsigned int drumstat;
static unsigned long chanage[18];
static unsigned char inst[256][11], chanfreq[18], chantrack[18];
static unsigned char trinst[16], trquant[16], trchan[16], trprio[16];
static unsigned char trvol[16], adlibchanoffs[9] = {0,1,2,8,9,10,16,17,18};
static unsigned int adlibfreq[63] =
{
	0,
	2390,2411,2434,2456,2480,2506,2533,2562,2592,2625,2659,2695,
	3414,3435,3458,3480,3504,3530,3557,3586,3616,3649,3683,3719,
	4438,4459,4482,4504,4528,4554,4581,4610,4640,4673,4707,4743,
	5462,5483,5506,5528,5552,5578,5605,5634,5664,5697,5731,5767,
	6486,6507,6530,6552,6576,6602,6629,6658,6688,6721,6755,6791,
	7510
};

int main (int argc, char **argv)
{
	float f;
	long i, j, k, track, volevel, templong, minicnt = 0, playtype = 1;
	long leng, daseconds, fil, loopnum, lastsample;
	short mousx, mousy, obstatus, bstatus;
	char ch, ksmfil[80], dumpfil[80], buffer[80];

	if (argc < 2)
	{
		printf("DigiKSM [KSM file][options]                               by Kenneth Silverman\n");
		printf("Options summary:\n");
		printf("   ?.WAV     - Dump digital music into a wave file\n");
		printf("   r#####    - Sample rate 6000-48000 (44100 default)\n");
		printf("   m/s       - Mono / Stereo (Stereo default - instruments in random places)\n");
		printf("   16/8      - Bits per Sample (16 default)\n");
		printf("   t###      - .KSM Tempo (60-960, 240 default)\n");
		printf("   l###      - Loop # times before quitting\n");
		printf("Example: >digiksm powsong powsong.wav r48000 16 t360\n");
		exit(0);
	}

	setbuf(stdout,0);            //Force printf's to update immediately

	samplerate = 44100;
	numspeakers = 2;
	bytespersample = 2;
	strcpy(ksmfil,argv[1]);
	dumpfil[0] = 0;
	speed = 240;
	loopnum = 0;

	i = 2;
	while (i < argc)
	{
		strcpy(buffer,argv[i]);

		if (strchr(buffer,'.') != 0)
			strcpy(dumpfil,buffer);
		else
		{
			j = 0;
			while (buffer[j] == 32) j++;

			if ((buffer[j] == 'r') || (buffer[j] == 'R'))
			{
				j++;
				samplerate = 0;
				while ((buffer[j] >= 48) && (buffer[j] <= 57))
				{
					samplerate *= 10;
					samplerate += buffer[j]-48;
					j++;
				}
			}
			else if ((buffer[j] == 't') || (buffer[j] == 'T'))
			{
				j++;
				speed = 0;
				while ((buffer[j] >= 48) && (buffer[j] <= 57))
				{
					speed *= 10;
					speed += buffer[j]-48;
					j++;
				}
				speed = min(max(speed,60),960);
			}
			else if ((buffer[j] == 'l') || (buffer[j] == 'L'))
			{
				j++;
				loopnum = 0;
				while ((buffer[j] >= 48) && (buffer[j] <= 57))
				{
					loopnum *= 10;
					loopnum += buffer[j]-48;
					j++;
				}
			}
			else if ((buffer[j] == 'm') || (buffer[j] == 'M')) numspeakers = 1;
			else if ((buffer[j] == 's') || (buffer[j] == 'S')) numspeakers = 2;
			else if (buffer[j] == '8') bytespersample = 1;
			else if ((buffer[j] == '1') && (buffer[j+1] == '6')) bytespersample = 2;
		}
		i++;
	}

	if (dumpfil[0] != 0)
	{
		if (loopnum <= 0) loopnum = 1;

		if ((fil = open(dumpfil,O_BINARY|O_CREAT|O_TRUNC|O_WRONLY,S_IWRITE)) >= 0)
		{
			write(fil,&buffer[0],44);

			adlibinit(samplerate,numspeakers,bytespersample);

			loadmusic(ksmfil);
			musicon();
			srand(0); randoinsts();
			printf("%5ldhz %c %ldbit Saving %s\n",samplerate,71+numspeakers*6,bytespersample<<3,ksmfil);
						
			leng = 0L; daseconds = 0;
			while (1)
			{
				k = nownote;
				ksmnexttic();
				if (nownote < k) { loopnum--; if (loopnum <= 0) break; }

				j = (samplerate*numspeakers*bytespersample/speed)&~3;
				adlibgetsample(sbbuf,j);
				if ((leng += write(fil,sbbuf,j)) < j)
					{ printf("\nWhoops!  Ran out of disk space.\n"); break; }
				else if (leng >= daseconds*samplerate*numspeakers*bytespersample)
					{ daseconds++; printf("%ld ",daseconds); }

				if (kbhit())
				{
					ch = getch();
					if (ch == 27) break;
				}
			}
			musicoff();

			buffer[0] = 'R'; buffer[1] = 'I'; buffer[2] = 'F'; buffer[3] = 'F';
			i = leng+36;
			buffer[4] = (i&255); buffer[5] = ((i>>8)&255); buffer[6] = ((i>>16)&255); buffer[7] = ((i>>24)&255);
			buffer[8] = 'W'; buffer[9] = 'A'; buffer[10] = 'V'; buffer[11] = 'E';
			buffer[12] = 'f'; buffer[13] = 'm'; buffer[14] = 't'; buffer[15] = ' ';
			buffer[16] = 16; buffer[17] = 0; buffer[18] = 0; buffer[19] = 0;
			buffer[20] = 1; buffer[21] = 0;
			buffer[22] = numspeakers; buffer[23] = 0;
			i = samplerate;
			buffer[24] = (i&255); buffer[25] = ((i>>8)&255); buffer[26] = ((i>>16)&255); buffer[27] = ((i>>24)&255);
			i = (samplerate<<(numspeakers+bytespersample-2));
			buffer[28] = (i&255); buffer[29] = ((i>>8)&255); buffer[30] = ((i>>16)&255); buffer[31] = ((i>>24)&255);
			buffer[32] = numspeakers*bytespersample; buffer[33] = 0;
			buffer[34] = (bytespersample<<3); buffer[35] = 0;
			buffer[36] = 'd'; buffer[37] = 'a'; buffer[38] = 't'; buffer[39] = 'a';
			i = leng;
			buffer[40] = (i&255); buffer[41] = ((i>>8)&255); buffer[42] = ((i>>16)&255); buffer[43] = ((i>>24)&255);
			lseek(fil,0L,SEEK_SET);
			write(fil,&buffer[0],44);

			close(fil);

			exit(0);
		}
	}

	if (loopnum <= 0) loopnum = 0x7fffffff;

	initsb(2,samplerate,numspeakers,bytespersample,30);
	loadmusic(ksmfil);
	adlibinit(samplerate,numspeakers,bytespersample);

	musicon();
	srand(0); randoinsts();

	printf("        %5ldhz %c %ldbit Playing %s",samplerate,71+numspeakers*6,bytespersample<<3,ksmfil);

	sbend = SBBUFSIZ-halfbufbytes;
	
	t0 = gettime586();
	while (1)
	{
		t1 = gettime586();
		while (sbplc >= sbend);  //Most of the CPU's cycles get wasted here!
		calccycles += t1-t0; t0 = gettime586();
		freecycles += t0-t1;
		if (calccycles+freecycles >= 2147483648) { calccycles >>= 1; freecycles >>= 1; }

		if (!musicstatus)
			if (sbend >= lastsample+SBBUFSIZ) { uninitsb(); exit(0); }
		while (sbplc < sbend)
		{
			j = min(sbend-sbplc,(minicnt/speed+4)&~3);
			if ((sbplc&(SBBUFSIZ-1)) < ((sbplc+j-1)&(SBBUFSIZ-1)))
				adlibgetsample(&sbbuf[sbplc&(SBBUFSIZ-1)],j);
			else
			{
				i = SBBUFSIZ-(sbplc&(SBBUFSIZ-1));
				adlibgetsample(&sbbuf[sbplc&(SBBUFSIZ-1)],i);
				adlibgetsample(&sbbuf[0],j-i);
			}
			sbplc += j; minicnt -= speed*j;
			while (minicnt <= 0)
			{
				minicnt += samplerate*numspeakers*bytespersample;

				k = nownote;
				ksmnexttic();
				if (nownote < k)
				{
					loopnum--;
					if (loopnum <= 0) { musicoff(); lastsample = sbplc; }
				}
				
				if (!(count&15))
				{
					obstatus = bstatus;
					readmousebstatus(&bstatus);
					if (bstatus&1)
					{
						readmousexy(&mousx,&mousy);
						if ((mousx) && (obstatus&1))
						{
							changesamplerate((samplerate*(mousx+512))>>9);
							adlibinit(samplerate,numspeakers,bytespersample);
							printf("\r        %5dhz",samplerate);
						}
					}
					printf("\r%3.2f%% ",((float)calccycles*100.0)/((float)(calccycles+freecycles)));
					if (kbhit())
					{
						ch = getch();
						if (ch == ' ') randoinsts();
						if (ch == 27) { musicoff(); uninitsb(); exit(0); }
					}
				}
			}
		}
	}
	return(0);
}

void setinst (int chan, char *v)
{
	adlib0(chan+0xa0,0);
	adlib0(chan+0xb0,0);
	adlib0(chan+0xc0,v[10]);
	chan = adlibchanoffs[chan];
	adlib0(chan+0x20,v[5]);
	adlib0(chan+0x40,v[6]);
	adlib0(chan+0x60,v[7]);
	adlib0(chan+0x80,v[8]);
	adlib0(chan+0xe0,v[9]);
	adlib0(chan+0x23,v[0]);
	adlib0(chan+0x43,v[1]);
	adlib0(chan+0x63,v[2]);
	adlib0(chan+0x83,v[3]);
	adlib0(chan+0xe3,v[4]);
}

void ksmnexttic ()
{
	long i, j, k, track, volevel, templong;

	count++;
	while (count >= countstop)
	{
		templong = note[nownote]; track = ((templong>>8)&15);
		if (!(templong&192))
		{
			i = 0;
			while (((chanfreq[i] != (templong&63)) || (chantrack[i] != ((templong>>8)&15))) && (i < numchans))
				i++;
			if (i < numchans)
			{
				adlib0(i+0xb0,(adlibfreq[templong&63]>>8)&0xdf);
				chanfreq[i] = chanage[i] = 0;
			}
		}
		else
		{
			volevel = trvol[track]^63;
			if ((templong&192) == 128) volevel = min(volevel+4,63);
			if ((templong&192) == 192) volevel = max(volevel-4,0);
			if (track < 11)
			{
				i = numchans; k = 0;
				for(j=0;j<numchans;j++)
					if ((countstop-chanage[j] >= k) && (chantrack[j] == track))
						{ k = countstop-chanage[j]; i = j; }
				if (i < numchans)
				{
					adlib0(i+0xb0,0);
					adlib0(adlibchanoffs[i]+0x43,(inst[trinst[track]][1]&192)+volevel);
					adlib0(i+0xa0,adlibfreq[templong&63]);
					adlib0(i+0xb0,(adlibfreq[templong&63]>>8)|32);
					chanfreq[i] = templong&63; chanage[i] = countstop;
				}
			}
			else if (drumstat&32)
			{
				j = adlibfreq[templong&63];
				switch(track)
				{
					case 11: i = 6; j -= 2048; break;
					case 12: case 15: i = 7; j -= 2048; break;
					case 13: case 14: i = 8; break;
				}
				adlib0(i+0xa0,j);
				adlib0(i+0xb0,(j>>8)&0xdf);
				adlib0(0xbd,drumstat&(~(32768>>track)));
				if ((track == 11) || (track == 12) || (track == 14))
					adlib0(adlibchanoffs[i]+0x43,(inst[trinst[track]][1]&192)+volevel);
				else
					adlib0(adlibchanoffs[i]+0x40,(inst[trinst[track]][6]&192)+volevel);
				drumstat |= (32768>>track);
				adlib0(0xbd,drumstat);
			}
		}
		nownote++; if (nownote >= numnotes) nownote = 0;
		templong = note[nownote];
		if (!nownote) count = (templong>>12)-1;
		i = 240 / trquant[(templong>>8)&15];
		countstop = (templong>>12)+(i>>1); countstop -= countstop%i;
	}
}

void loadmusic (char *filename)
{
	char buffer[256], instbuf[11];
	int infile, i, j;

	if (firstime == 1)
	{
		if ((infile = open("insts.dat",O_BINARY|O_RDONLY,S_IREAD)) == -1)
			{ printf("I cannot load insts.dat!"); exit(0); }
		for(i=0;i<256;i++)
		{
			read(infile,buffer,33);
			for(j=0;j<11;j++) inst[i][j] = buffer[j+20];
		}
		close(infile);
		numchans = 9;
		adlib0(0x1,32);  //clear test stuff
		adlib0(0x4,0);   //reset
		adlib0(0x8,0);   //2-operator synthesis
		firstime = 0;
	}
	strcpy(buffer,filename);
	if ((!strstr(buffer,".KSM")) && (!strstr(buffer,".ksm"))) strcat(buffer,".KSM");
	if ((infile = open(buffer,O_BINARY|O_RDONLY,S_IREAD)) == -1)
		{ printf("I cannot load %s.",filename); exit(0); }
	read(infile,trinst,16);
	read(infile,trquant,16);
	read(infile,trchan,16);
	read(infile,trprio,16);
	read(infile,trvol,16);
	read(infile,&numnotes,2);
	read(infile,note,numnotes<<2);
	close(infile);
	if (!trchan[11]) { drumstat = 0; numchans = 9; adlib0(0xbd,drumstat); }
	if (trchan[11] == 1)
	{
		for(i=0;i<11;i++) instbuf[i] = inst[trinst[11]][i];
		instbuf[1] = ((instbuf[1]&192)|(trvol[11])^63);
		setinst(6,instbuf);
		for(i=0;i<5;i++) instbuf[i] = inst[trinst[12]][i];
		for(i=5;i<11;i++) instbuf[i] = inst[trinst[15]][i];
		instbuf[1] = ((instbuf[1]&192)|(trvol[12])^63);
		instbuf[6] = ((instbuf[6]&192)|(trvol[15])^63);
		setinst(7,instbuf);
		for(i=0;i<5;i++) instbuf[i] = inst[trinst[14]][i];
		for(i=5;i<11;i++) instbuf[i] = inst[trinst[13]][i];
		instbuf[1] = ((instbuf[1]&192)|(trvol[14])^63);
		instbuf[6] = ((instbuf[6]&192)|(trvol[13])^63);
		setinst(8,instbuf);
		drumstat = 32; numchans = 6; adlib0(0xbd,drumstat);
	}
}

void musicon ()
{
	long i, j, k;
	char instbuf[11];

	for(i=0;i<numchans;i++) chantrack[i] = chanage[i] = 0;
	j = 0;
	for(i=0;i<16;i++)
		if ((trchan[i] > 0) && (j < numchans))
		{
			k = trchan[i];
			while ((j < numchans) && (k > 0)) { chantrack[j] = i; k--; j++; }
		}
	for(i=0;i<numchans;i++)
	{
		for(j=0;j<11;j++) instbuf[j] = inst[trinst[chantrack[i]]][j];
		instbuf[1] = ((instbuf[1]&192)|(trvol[chantrack[i]]^63));
		setinst(i,instbuf);
		chanfreq[i] = 0;
	}
	count = countstop = (note[0]>>12)-1;
	nownote = 0; musicstatus = 1;
}

void musicoff ()
{
	int i;

	for(i=0;i<numchans;i++) { adlib0(0xa0+i,0); adlib0(0xb0+i,0); }
	musicstatus = 0;
}

void randoinsts ()
{
	long i, j, k;
	float f;

	if (numspeakers == 2)
	{
		j = (rand()&2)-1; k = 0;
		for(i=0;i<9;i++)
		{
			if ((i == 0) || (chantrack[i] != chantrack[i-1]))
			{
				f = (float)rand()/32768.0;
				if (j > 0)
				{
						//lvol[i] < rvol[i]
					lvol[i] = log2(f+1);
					rvol[i] = log2(3-f);
					lplc[i] = rand()&255;
					rplc[i] = 0;
				}
				else
				{
						//lvol[i] > rvol[i]
					lvol[i] = log2(3-f);
					rvol[i] = log2(f+1);
					lplc[i] = 0;
					rplc[i] = rand()&255;
				}
				j = -j;
				if (((drumstat&32) == 0) || (i < 6)) k++;
			}
			else
			{
				lvol[i] = lvol[i-1]; rvol[i] = rvol[i-1];
				lplc[i] = lplc[i-1]; rplc[i] = rplc[i-1];
			}
		}
		if (k < 2)  //If only 1 source, force it to be in center
		{
			if (drumstat&32) i = 5; else i = 8;
			for(;i>=0;i--)
			{
				lvol[i] = rvol[i] = 1;
				lplc[i] = rplc[i] = 0;
			}
		}
	}
}
