//SBSTREAM.C by Ken Silverman (http://advsys.net/ken)
#include <memory.h>
#include <conio.h>
#include <stdlib.h>
#include <stdio.h>
#include <dos.h>

	//FIFO variables
#define SBBUFSIZ 16384
char sbbuf[SBBUFSIZ];
volatile long sbplc = 0, sbend = 0;

	//Actual sound parameters after initsb was called
long samplerate, numspeakers, bytespersample, intspersec;

static char alreadyinited = 0, sbmode = 0; //0=off, 1=record, 2=play
long sbport = 0x220, sbirq = 0x7, sbdma8 = 0x1, sbdma16 = 0x1, sbdma, sbver;
char dmapagenum[8] = {0x87,0x83,0x81,0x82,0x8f,0x8b,0x89,0x8a};
static unsigned short sndseg = 0, sndselector;
static long kdminprep = 0, mixerval = 0;
static long dmacheckport, dmachecksiz;
volatile long sndoffs, sndoffsplc, sndoffsxor;
long bytespertic, sndbufsiz, halfbufbytes;

void (__interrupt __far *oldsbhandler)();
void __interrupt __far sbhandler();

unsigned short convallocate (unsigned short);
#pragma aux convallocate =\
	"mov ax, 0x100"\
	"int 0x31"\
	"jnc noerror"\
	"mov ax, 0"\
	"noerror: mov sndselector, dx"\
	parm [bx] modify [eax edx]

void convdeallocate (unsigned short);
#pragma aux convdeallocate =\
	"mov ax, 0x101"\
	"mov dx, sndselector"\
	"int 0x31"\
	parm [dx] modify [eax edx]

void resetsb ();
#pragma aux resetsb =\
	"mov edx, sbport"\
	"add edx, 0x6"\
	"mov al, 1"\
	"out dx, al"\
	"xor al, al"\
	"delayreset: dec al"\
	"jnz delayreset"\
	"out dx, al"\
	"mov ecx, 65536"\
	"empty: mov edx, sbport"\
	"add edx, 0xe"\
	"in al, dx"\
	"test al, al"\
	"jns nextattempt"\
	"sub dl, 4"\
	"in al, dx"\
	"cmp al, 0aah"\
	"je resetok"\
	"dec ecx"\
	"nextattempt: jnz empty"\
	"mov eax, -1"\
	"jmp resetend"\
	"resetok: xor eax, eax"\
	"resetend:"\
	modify [ebx ecx edx]

long sbin ();
#pragma aux sbin =\
	"xor eax, eax"\
	"mov edx, sbport"\
	"add edx, 0xe"\
	"busy: in al, dx"\
	"or al, al"\
	"jns busy"\
	"sub edx, 4"\
	"in al, dx"\
	modify [edx]

void sbout (long);
#pragma aux sbout =\
	"mov edx, sbport"\
	"add edx, 0xc"\
	"shl eax, 8"\
	"busy: in al, dx"\
	"or al, al"\
	"js busy"\
	"shr eax, 8"\
	"out dx, al"\
	parm [eax] modify [edx]

long sbmixin (long);
#pragma aux sbmixin =\
	"mov edx, sbport"\
	"add edx, 0x4"\
	"out dx, al"\
	"inc edx"\
	"xor eax, eax"\
	"in al, dx"\
	parm [eax] modify [edx]

long sbmixout (long, long);
#pragma aux sbmixout =\
	"mov edx, sbport"\
	"add edx, 0x4"\
	"out dx, al"\
	"inc edx"\
	"mov eax, ebx"\
	"out dx, al"\
	parm [eax][ebx] modify [edx]

void koutp (long, long);
#pragma aux koutp =\
	"out dx, al"\
	parm [edx][eax]

long kinp (long);
#pragma aux kinp =\
	"xor eax, eax"\
	"in al, dx"\
	parm [edx]

void initsb (char dasbmode, long dasamplerate, char danumspeakers, char dabytespersample, char daintspersec)
{
	long i, j, k;

	if (sbmode) uninitsb();
	sbmode = dasbmode;
	sbplc = sbend = 0;

	if (!alreadyinited) getsbset();
	resetsb();
	if (!alreadyinited) { sbout(0xe1); sbver = (sbin()<<8); sbver += sbin(); }

	samplerate = min(max(dasamplerate,6000),48000);
	numspeakers = min(max(danumspeakers,1),(sbver>=0x0300)+1);
	bytespersample = min(max(dabytespersample,1),(sbver>=0x0400)+1);
	intspersec = min(max(daintspersec,1),120);
	if (sbmode == 1)  //record
	{
		if (sbver < 0x0201) samplerate = min(samplerate,13000);
		if (sbver < 0x0300) samplerate = min(samplerate,15000);
	}
	else              //play
	{
		if (sbver < 0x0201) samplerate = min(samplerate,22050);
		memset(sbbuf,(bytespersample-2)&128,SBBUFSIZ);  //Silence (0 or 128)
	}
		//SBPRO preference of 44100m over 22050s
	if ((sbver < 0x0400) && (samplerate > 22050)) numspeakers = 1;

	if (bytespersample == 2) sbdma = sbdma16; else sbdma = sbdma8;

	bytespertic = (((samplerate/120)+1)&~1);
	sndbufsiz = bytespertic*(120/intspersec);
	halfbufbytes = (sndbufsiz<<(bytespersample+numspeakers-2));

	if (!sndseg)    //Allocate DMA buffer in conventional memory
	{
		if (!(sndseg = convallocate(((halfbufbytes<<2)+15)>>4)))
			{ printf("Can't allocate conventional memory for sound\n"); exit(0); }
		sndoffs = (((long)sndseg)<<4);

		if ((sndoffs&65535)+(halfbufbytes<<1) >= 65536)   //64K DMA page check
			sndoffs += (halfbufbytes<<1);

		sndoffsplc = sndoffs;
		sndoffsxor = sndoffsplc ^ (sndoffsplc+halfbufbytes);

		memset((void *)sndoffs,0L,halfbufbytes<<1);
	}

	if (sbirq < 8)
	{
		oldsbhandler = _dos_getvect(sbirq+0x8);           //Set new IRQ7 vector
		_disable(); _dos_setvect(sbirq+0x8, sbhandler); _enable();
		koutp(0x21,kinp(0x21) & ~(1<<(sbirq&7)));
	}
	else
	{
		oldsbhandler = _dos_getvect(sbirq+0x68);    //Set new SB IRQ vector
		_disable(); _dos_setvect(sbirq+0x68, sbhandler); _enable();
		koutp(0xa1,kinp(0xa1) & ~(1<<(sbirq&7)));
	}

	if (sbdma < 4)  //Set up DMA REGISTERS
	{
		dmacheckport = (sbdma<<1)+1;
		dmachecksiz = (halfbufbytes<<1);
		koutp(0xa,sbdma+4);
		koutp(0xc,0);
		koutp(0xb,0x50+(sbmode<<2)+sbdma);      //autoinit: 0x54-rec, 0x58-play
		koutp(sbdma<<1,sndoffs&255);
		koutp(sbdma<<1,(sndoffs>>8)&255);
		koutp(dmacheckport,(dmachecksiz-1)&255);
		koutp(dmacheckport,((dmachecksiz-1)>>8)&255);
		koutp(dmapagenum[sbdma],((sndoffs>>16)&255));
		koutp(0xa,sbdma);
	}
	else
	{
		dmacheckport = ((sbdma-4)<<2)+0xc2;
		dmachecksiz = halfbufbytes;
		koutp(0xd4,sbdma);
		koutp(0xd8,0);
		koutp(0xd6,0x50+(sbmode<<2)+(sbdma-4)); //autoinit: 0x54-rec, 0x58-play
		koutp(dmacheckport-2,(sndoffs>>1)&255);
		koutp(dmacheckport-2,((sndoffs>>1)>>8)&255);
		koutp(dmacheckport,(dmachecksiz-1)&255);
		koutp(dmacheckport,((dmachecksiz-1)>>8)&255);
		koutp(dmapagenum[sbdma],((sndoffs>>16)&255));
		koutp(0xd4,sbdma-4);
	}

	sbout(0xd5-(sbmode<<1));   //SB Speaker: 0xd3:rec(off), 0xd1:play(on)

	if (sbver < 0x0400)
	{
		sbout(0x40);                    //SB Set speed
		sbout(256-(1000000/(samplerate<<(numspeakers-1))));
	}

	if (sbver < 0x0200)
	{
		sbout(0x34-(sbmode<<4));        //SB 1-shot mode: 0x24:rec, 0x14:play
		sbout((sndbufsiz-1)&255);
		sbout(((sndbufsiz-1)>>8)&255);
	}
	else
	{
		if (sbver < 0x0400)
		{
				//Set length for auto-init mode
			sbout(0x48);
			sbout((halfbufbytes-1)&255);
			sbout(((halfbufbytes-1)>>8)&255);
			if (sbmode == 1)
			{
				if (numspeakers == 2)                   //SB set stereo
				{
					sbmixout(0L,0L);
					sbmixout(0xe,0xa8);
				}
				if ((samplerate<<(numspeakers-1)) <= 22050)
					sbout(0x2c);                         //Start SB interrupts!
				else
					sbout(0x98);                         //High Speed DMA
			}
			else
			{
				if (numspeakers == 2)                   //SB set stereo
				{
					sbmixout(0L,0L);
					sbmixout(0xe,sbmixin(0xe)|2);
				}
				if ((samplerate<<(numspeakers-1)) <= 22050)
					sbout(0x1c);                         //Start SB interrupts!
				else
					sbout(0x90);                         //High Speed DMA
			}
		}
		else
		{
			sbout(0x43-sbmode);
			sbout((samplerate>>8)&255);
			sbout(samplerate&255);

			sbout(0xd6-(sbmode<<3)-((bytespersample-1)<<4));
			sbout(((bytespersample-1)<<4)+((numspeakers-1)<<5));
			i = (sndbufsiz<<(numspeakers-1))-1;
			sbout(i&255);
			sbout((i>>8)&255);
		}
	}
	alreadyinited = 1;
}

void getsbset ()
{
	char *sbset;
	long i;

	sbset = getenv("BLASTER");

	i = 0;
	while (sbset[i] > 0)
	{
		switch(sbset[i])
		{
			case 'A': case 'a':
				i++;
				sbport = 0;
				while (((sbset[i] >= 48) && (sbset[i] <= 57)) ||
						 ((sbset[i] >= 'A') && (sbset[i] <= 'F')) ||
						 ((sbset[i] >= 'a') && (sbset[i] <= 'f')))
				{
					sbport *= 16;
					if ((sbset[i] >= 48) && (sbset[i] <= 57)) sbport += sbset[i]-48;
					if ((sbset[i] >= 'A') && (sbset[i] <= 'F')) sbport += sbset[i]-55;
					if ((sbset[i] >= 'a') && (sbset[i] <= 'f')) sbport += sbset[i]-55-32;
					i++;
				}
				break;
			case 'I': case 'i':
				i++;
				sbirq = 0;
				while ((sbset[i] >= 48) && (sbset[i] <= 57))
				{
					sbirq *= 10;
					sbirq += sbset[i]-48;
					i++;
				}
				break;
			case 'D': case 'd':
				i++;
				sbdma8 = 0;
				while ((sbset[i] >= 48) && (sbset[i] <= 57))
				{
					sbdma8 *= 10;
					sbdma8 += sbset[i]-48;
					i++;
				}
				break;
			case 'H': case 'h':
				i++;
				sbdma16 = 0;
				while ((sbset[i] >= 48) && (sbset[i] <= 57))
				{
					sbdma16 *= 10;
					sbdma16 += sbset[i]-48;
					i++;
				}
				break;
			default:
				i++;
				break;
		}
	}
}

void changesamplerate (long newsamplerate)
{
	if (!sbmode) return;

	newsamplerate = min(max(newsamplerate,6000),48000);
	if (sbver < 0x0400)
	{
			//Doesn't work on SBPCI128 (SBPRO compatible mode)
		/*if (sbmode == 1)  //record
		{
			if (sbver < 0x0201) newsamplerate = min(newsamplerate,13000);
			if (sbver < 0x0300) newsamplerate = min(newsamplerate,15000);
		}
		else              //play
		{
			if (sbver < 0x0201) newsamplerate = min(newsamplerate,22050);
		}
		if (newsamplerate != samplerate)
		{
			samplerate = newsamplerate;
			sbout(0xd0); //DSP Halt
			sbout(0x40);
			sbout(256-(1000000/(samplerate<<(numspeakers-1))));
			sbout(0xd4); //DSP Continue
		}*/
	}
	else if (newsamplerate != samplerate)
	{
		samplerate = newsamplerate;
		sbout(0x43-sbmode);
		sbout((samplerate>>8)&255);
		sbout(samplerate&255);
		sbout(0xd6); //16-bit DSP Continue
	}
}

void __interrupt __far sbhandler()
{
	long i;

	//if (sbirq == 7)
	//{
	//   outp(0x20,0xb);         //select PIC In Service Register
	//   if (!(inp(0x20)&128))   //read In Service Register - test for true IRQ7
	//      { koutp(0x20,0x20); return; }
	//}

	if (sbver < 0x0200)
	{
		sbout(0x34-(sbmode<<4));        //SB 1-shot mode: 0x24:rec, 0x14:play
		sbout((sndbufsiz-1)&255);
		sbout(((sndbufsiz-1)>>8)&255);
		kinp(sbport+0xe);                      //Acknowledge SB
	}
	else
	{
		mixerval = sbmixin(0x82);
		if (mixerval&1) kinp(sbport+0xe);      //Acknowledge 8-bit DMA
		if (mixerval&2) kinp(sbport+0xf);      //Acknowledge 16-bit DMA
	}
	if (sbirq >= 8) koutp(0xa0,0x20);
	koutp(0x20,0x20);

	if (kdminprep != 0) return;

	i = kinp(dmacheckport); i += (kinp(dmacheckport)<<8);
	if (i < dmachecksiz)
		if ((sndoffsplc<(sndoffsplc^sndoffsxor)) == (i<(dmachecksiz>>1)))
			sndoffsplc ^= sndoffsxor;

	_enable();

	kdminprep = 1;

	sndoffsplc ^= sndoffsxor;
	i = SBBUFSIZ-(sbend&(SBBUFSIZ-1));  //i = room left in sbbuf

	if (sbmode == 1)
	{
		if (halfbufbytes <= i)
			memcpy((void *)&sbbuf[sbend&(SBBUFSIZ-1)],(void *)sndoffsplc,halfbufbytes);
		else
		{
			memcpy((void *)&sbbuf[sbend&(SBBUFSIZ-1)],(void *)sndoffsplc,i);
			memcpy((void *)sbbuf,(void *)(sndoffsplc+i),halfbufbytes-i);
		}
	}
	else
	{
		if (halfbufbytes <= i)
			memcpy((void *)sndoffsplc,(void *)&sbbuf[sbend&(SBBUFSIZ-1)],halfbufbytes);
		else
		{
			memcpy((void *)sndoffsplc,(void *)&sbbuf[sbend&(SBBUFSIZ-1)],i);
			memcpy((void *)(sndoffsplc+i),(void *)sbbuf,halfbufbytes-i);
		}
	}

	sbend += halfbufbytes;

	kdminprep = 0;
}

void uninitsb ()
{
	if (!sbmode) return;
	sbmode = 0;

		//Mask off DMA
	if (sbdma < 4) koutp(0xa,sbdma+4); else koutp(0xd4,sbdma);

	if (sbver >= 0x0400) sbout(0xda-(bytespersample-1));
	resetsb();
	sbout(0xd3);                           //Turn speaker off

	if (sndseg != 0) convdeallocate(sndseg), sndseg = 0;

	if (sbirq < 8)
	{
		koutp(0x21,kinp(0x21) | (1<<(sbirq&7)));
		_disable(); _dos_setvect(sbirq+0x8, oldsbhandler); _enable();
	}
	else
	{
		koutp(0xa1,kinp(0xa1) | (1<<(sbirq&7)));
		_disable(); _dos_setvect(sbirq+0x68, oldsbhandler); _enable();
	}
}
